<?php

namespace Nwb\UrlToPdf;

class UrlToPdf
{
    /**
     * @var string
     */
    private string $apiUrl = 'https://url-to-pdf.kub.nwb.fr';

    /**
     * @var string|null
     */
    private ?string $accessKey;

    /**
     * UrlToPdf constructor.
     * @param array $options Liste des options de configuration
     * @throws Exception
     */
    public function __construct(array $options = [])
    {
        $key = $options['accessKey'] ?? $_ENV['URL_TO_PDF_ACCESS_KEY'] ??  null;

        if (!$key) {
            throw new \Exception('Missing access key');
        }

        $this->accessKey = $key;

        if (!empty($options['apiUrl'])) {
            $this->apiUrl = $options['apiUrl'];
        }
    }

    /**
     * Requête curl vers l'API
     * 
     * @param string $url contient la fin de l'url pour appeler la bonne méthode de l'API (ex : /get-pdf)
     * @param array $options Liste des options de configuration, notamment $options['url'] l'url vers la page html à convertir en PDF
     * @return false|resource contenu du PDF
     * @throws Exception si la page est vide
     * @throws Exception si le code HTTP n'est pas 200
     * @throws Exception si le contenu de la réponse n'est pas un PDF
     * @throws Exception si le contenu de la réponse est vide
     */
    private function request(string $url, array $options)
    {
        // Check options
        if (empty($options['url'])) {
            throw new \Exception('Missing url');
        }
        if (!empty($options['auth']) && !empty($options['auth']['username']) && !empty($options['auth']['password'])) {
            $username = $options['auth']['username'] ?? null;
            $password = $options['auth']['password'] ?? null;
        }
        !empty($options['loadingOptions']) ? $options['loadingOptions'] = array_push($options['loadingOptions'], ['credentials' => ['username' => $username, 'password' => $password]]) : $options['loadingOptions'] = ['credentials' => ['username' => $username, 'password' => $password]];

        // Init 1er curl pour récupérer le contenu de la page html
        $ch1 = curl_init();
        curl_setopt($ch1, CURLOPT_URL, $options['url']);
        curl_setopt($ch1, CURLOPT_RETURNTRANSFER, true);
        // Si options d'authentifications on les ajoute
        if (!empty($options['auth'])) {
            curl_setopt($ch1, CURLOPT_USERPWD, $options['auth']['username'] . ':' . $options['auth']['password']);
        }
        $pageContent = curl_exec($ch1);

        // Erreur cURL
        if ($pageContent === false) {
            throw new \Exception("Erreur cURL lors de la récupération du fichier html : " . curl_error($ch1));
        }

        // Vérifier si la page est vide
        if (empty($pageContent)) {
            throw new \Exception("La page html récupérée n'a pas de contenu.");
        }

        // Vérifier si le code HTTP est 200
        $httpCode = curl_getinfo($ch1, CURLINFO_HTTP_CODE);
        $contentType = curl_getinfo($ch1, CURLINFO_CONTENT_TYPE);
        if ($httpCode !== 200) {
            if ($pageContent) {
                throw new \Exception("Une erreur s'est produite lors de la vérification du contenu du fichier html : " . $pageContent);
            }

            throw new \Exception('Error: ' . $httpCode);
        }

        curl_close($ch1);

        // Init 2ème curl pour appeler l'API
        $ch = curl_init();

        // Définition des headers
        $headers = [
            'Content-Type: application/json',
            'ACCESS-KEY: ' . $this->accessKey,
        ];

        curl_setopt($ch, CURLOPT_URL, $this->apiUrl . $url);
        curl_setopt($ch, CURLOPT_POST, true);
        curl_setopt($ch, CURLOPT_POSTFIELDS, json_encode($options));
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        $response = curl_exec($ch);

        // Erreur cURL
        if ($response === false) {
            throw new \Exception("Erreur cURL : " . curl_error($ch));
        }

        // Vérifier si la réponse est vide
        $httpCode = curl_getinfo($ch, CURLINFO_HTTP_CODE);
        $contentType = curl_getinfo($ch, CURLINFO_CONTENT_TYPE);
        if ($httpCode !== 200) {
            if (strpos($contentType, 'application/json') !== false) {
                $errorText = json_decode($response, true);

                if (count($errorText)) {
                    throw new \Exception($errorText);
                }
            } else {
                if ($response) {
                    throw new \Exception($response);
                }
            }

            throw new \Exception('Error: ' . $httpCode);
        }

        if (empty($response) || substr($response, 0, 4) !== "%PDF") {
            throw new \Exception('La réponse retournée est vide ou n\'est pas un PDF.');
        }

        return $response;
    }

    /**
     * Enregistre le contenu dans un fichier sur le disque.
     *
     * @param string $content Contenu à enregistrer.
     * @param string $destination Chemin de destination du fichier.
     * @return bool
     */
    public function saveTo(string $content, string $destination): bool
    {
        // Écriture du contenu dans le fichier
        return file_put_contents($destination, $content) !== false;
    }

    /**
     * Retourne un PDF à partir d'une URL et des options
     * 
     * @param array $options Liste des options de configuration
     * @return false|resource contenu du PDF
     */
    public function getPdf(?array $options)
    {
        $options = $options ?? [
            'url' => 'https://normandiewebschool.fr/cursus/bachelor-developpement-web/',
            'loadingOptions' => ['waitUntil' => 'networkidle2'],
            'pdfOptions' => ['format' => 'a4']
        ];

        return $this->request('/get-pdf', $options);
    }

    /**
     * Retourne un PDF à partir d'un contenu HTML
     * 
     * @param string $htmlContent type TCPDF qui peut contenir des liens vers des images/fichiers sur le disque (ex: src:"/var/www/html/public-data/tmp/url_to_pdf/mon_image.jpg")
     * @param string $projectRootPath Chemin racine du projet (ex: /var/www/html/)
     * @param string $webPath Chemin vers le répertoire web (ex: /var/www/html/public-data/tmp/url_to_pdf/)
     * @param string $webPathUrl Chemin vers le répertoire web accessible depuis l'API (ex: https://www.nwb.fr/public-data/tmp/url_to_pdf/)
     * @param array $pdfOptions Liste des options de configuration (marges, format, etc.). Par défaut : ['format' => 'a4', 'printBackground' => true, 'margin' => ['top' => '0.5cm', 'right' => '0.5cm', 'bottom' => '0.5cm', 'left' => '0.5cm']]
     * @param array $authOptions Liste des options d'authentification (username, password). Utile pour une preprod par exemple.
     * @param bool $isTestMode Indique si on est en mode test, dans ce cas on utilise l'API pour envoyer le fichier dans l'url en Base64
     * @return false|resource chemin du PDF
     * @throws Exception
     * 
     */
    public function getPdfFromHtml(string $htmlContent, string $projectRootPath, string $webPath, string $webPathUrl, array $pdfOptions = [], array $authOptions = [], bool $isTestMode = false): string
    {
        // Si répertoire web n'existe pas on trhow une exception
        if (!is_dir($webPath)) {
            throw new \Exception("Le répertoire web n'existe pas : " . $webPath);
        }

        // Test si $webPathUrl existe et fait plus de 8 caractères, si non on throw une exception
        if (!$webPathUrl || strlen($webPathUrl) < 8) {
            throw new \Exception("Le chemin web n'est pas valide : " . $webPathUrl);
        }

        // Test si $webPath se termine par un slash, si non on le rajoute
        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        // Test si $webPathUrl se termine par un slash, si non on le rajoute
        if (substr($webPathUrl, -1) !== '/') {
            $webPathUrl .= '/';
        }

        // Test si $webPathUrl ne commence pas par https:// ou http:// , si non on le rajoute
        $webPathUrl = trim($webPathUrl);
        if (strpos($webPathUrl, 'https://', 0) !== 0 && strpos($webPathUrl, 'http://', 0) !== 0) {
            $webPathUrl = ($isTestMode) ? 'http://' . $webPathUrl : 'https://' . $webPathUrl;
        }

        // Check options PDF et définition des options de base Puppetter par défaut
        $pdfOptions = $pdfOptions ?? [
            'format' => 'a4',
        ];

        // On check les options dédiées au module
        $pdfOptions['noBreakTableRows'] = $pdfOptions['noBreakTableRows'] ?? true;

        // Adapte le html pour accessibilité web 
        $htmlContent = $this->getCleanHtml($htmlContent, $projectRootPath, $webPath, $webPathUrl, $pdfOptions, $isTestMode);

        // On hash le contenu html pour créer un nom de fichier unique
        $fileName = md5($htmlContent) . '.html';

        // On créer un sous dossier pour stocker les PDF
        $webPath .= 'pdf/';

        // On créer le répertoire s'il n'existe pas
        if (!is_dir($webPath)) {
            mkdir($webPath, 0777, true);
        }

        // On créer le fichier
        $filePath = $webPath . $fileName;
        $file = fopen($filePath, 'w');

        // Enregistre le html dans un fichier temporaire
        fwrite($file, $htmlContent);
        fclose($file);

        // On met les droits en 777
        chmod($filePath, 0777);

        // Appel de l'API getPdf
        if (!$isTestMode) {
            $pdf = $this->getPdf([
                'url' => $webPathUrl . 'pdf/' . $fileName,
                'pdfOptions' => $pdfOptions,
                'auth' => $authOptions
            ]);
        }

        // Si on est en mode test on passe le contenu du fichier en base64 dans l'url (valable uniquement pour les petits fichiers)
        if ($isTestMode) {
            // On enregistre le html dans un fichier temporaire
            fopen($webPath . 'TEST_' . $fileName, 'w');
            file_put_contents($webPath . 'TEST_' . $fileName, $htmlContent);

            $htmlContent = base64_encode($htmlContent);

            // Troncature pour que le html content fasse moins de 3000 caractères (limite du GET)
            $htmlContent = substr($htmlContent, 0, 3000);

            // Appel de getPdf avec une url générée par l'API BASE64ToHtml
            $pdf = $this->getPdf([
                'url' => 'https://n8n.kub.nwb.fr/webhook/Base64ToHtml?htmlB64=' . $htmlContent,
                'pdfOptions' => $pdfOptions
            ]);
        }

        // Remplacement du contenu du fichier temporaire par le contenu du PDF et on renomme le fichier en .pdf
        $file = fopen($filePath, 'w');
        fwrite($file, $pdf);
        fclose($file);
        rename($filePath, $webPath . pathinfo($filePath, PATHINFO_FILENAME) . '.pdf');
        $filePath = $webPath . pathinfo($filePath, PATHINFO_FILENAME) . '.pdf';

        // On supprime les imafges temporaires sauf si on est en mode test
        if (!$isTestMode) {
            $this->deleteTempImages($webPath);
        }

        // Retourne le path temporaire du PDF
        return $filePath;
    }


    /**
     * Retourne le html passé en paramètre avec les path des fichiers accessibles pour l'API Url to PDF
     * 
     * @param string $htmlIn
     * @param string $projectRootPath (ex: /var/www/html/)
     * @param string $webPath (ex: /var/www/html/public-data/tmp/url_to_pdf/)
     * @param string $webPathUrl (ex: https://www.nwb.fr/public-data/tmp/url_to_pdf/)
     * @param array $pdfOptions Liste des options de configuration dédiées au module et à la conversion html
     * @param bool $isTestMode Indique si on est en mode test, dans ce cas on ne n'insère pas les path sous forme d'url
     * @return string $htmlContent
     * @throws Exception si l'image n'existe pas
     */
    public function getCleanHtml(string $htmlIn, string $projectRootPath, string $webPath, string $webPathUrl, array $pdfOptions = [], bool $isTestMode = false): string
    {
        if (!$htmlIn) {
            throw new \Exception("Le contenu HTML est vide.");
        }
        if (!$projectRootPath) {
            throw new \Exception("Le chemin racine du projet n'est pas renseigné.");
        }
        if (!$webPath) {
            throw new \Exception("Le chemin vers le répertoire web n'est pas renseigné.");
        }
        if (!$webPathUrl) {
            throw new \Exception("L'url vers le répertoire web n'est pas renseigné.");
        }

        // Test si $webPath se termine par un slash, si non on le rajoute
        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        // Test si $webPathUrl se termine par un slash, si non on le rajoute
        if (substr($webPathUrl, -1) !== '/') {
            $webPathUrl .= '/';
        }

        // On créer le répertoire s'il n'existe pas
        if (is_dir($webPath)) {
            $this->deleteDir($webPath);
        }
        mkdir($webPath, 0777, true);

        // On récupère le contenu HTML
        $htmlContent = $htmlIn;

        // On parse le contenu HTML
        $dom = new \DOMDocument();
        $dom->loadHTML($htmlContent);

        // On définit les tags et attributs à traiter
        $tagAttributes = [
            'img' => 'src',
            'script' => 'src',
            'link' => 'href',
            'page_header' => '',
            'page_footer' => '',
            'page' => '',
        ];

        // On parcourt tous les éléments à traiter
        foreach ($tagAttributes as $tag => $attribute) {
            $elements = $dom->getElementsByTagName($tag);
            if (empty($elements)) {
                continue;
            }

            foreach ($elements as $element) {
                // On traite d'abord les balises <page> <page_header> et <page_footer> provenant de la librairie TCPDF
                if ($tag === 'page' || $tag === 'page_header' || $tag === 'page_footer') {
                    // Conversion des balises en div 
                    $newElement = $dom->createElement('div');
                    $newElement->setAttribute('id', $tag);
                    $newElement->setAttribute('class', $tag);
                    if (!empty($element->getAttribute('style'))) {
                        $newElement->setAttribute('style', $element->getAttribute('style'));
                    }
                    // Si la balise à des attributs connus comme la gestion des marges on les convertit en style
                    $attributes = $element->attributes;
                    foreach ($attributes as $attrName => $attrNode) {
                        if ($attrName === 'backtop' || $attrName === 'backbottom' || $attrName === 'backleft' || $attrName === 'backright') {
                            switch ($attrName) {
                                case 'backtop':
                                    $newElement->setAttribute('style', $newElement->getAttribute('style') . 'margin-top:' . $attrNode->nodeValue . ';');
                                    break;
                                case 'backbottom':
                                    $newElement->setAttribute('style', $newElement->getAttribute('style') . 'margin-bottom:' . $attrNode->nodeValue . ';');
                                    break;
                                case 'backleft':
                                    $newElement->setAttribute('style', $newElement->getAttribute('style') . 'margin-left:' . $attrNode->nodeValue . ';');
                                    break;
                                case 'backright':
                                    $newElement->setAttribute('style', $newElement->getAttribute('style') . 'margin-right:' . $attrNode->nodeValue . ';');
                                    break;
                            }
                        }
                    }
                    // On déplace les enfants de la balise dans la nouvelle balise
                    while ($element->firstChild) {
                        $newElement->appendChild($element->firstChild);
                    }
                    $element->parentNode->replaceChild($newElement, $element);
                }

                $filePath = $element->getAttribute($attribute);

                // Test si le chemin est vide
                if (!$filePath) {
                    continue;
                }

                // Si le chemin contient ./ ou ../ ou // on les supprime
                $filePath = str_replace('../', '', $filePath);
                $filePath = str_replace('./', '', $filePath);


                // Si le chemin est une url ou un protocole de fichier ftp ou s3 on supprime la partie protocole et domaine
                $formatPath = "/^(https?|ftp|s3):\/\/(localhost|[\da-z\.-]+)(:[0-9]+)?(\.[a-z\.]{2,6})?([\/\w \.-]*)*\/?$/";
                if (preg_match($formatPath, $filePath)) {
                    $filePath = parse_url($filePath, PHP_URL_PATH);
                }

                // Si un '//' est présent on le supprime
                $filePath = str_replace('//', '/', $filePath);

                // Si le chemin commence par un slash et contient le pathRoot on le supprime
                if (strpos($filePath, '/', 0) === 0 && strpos($filePath, $projectRootPath, 0) !== 0) {
                    $filePath = substr($filePath, 1);
                }

                // Test du chemin réel
                if (realpath($filePath)) {
                    $filePath = realpath($filePath);
                }

                // On ajoute le chemin racine du projet s'il n'est pas déjà présent
                if (strpos($filePath, $projectRootPath, 0) !== 0) {
                    $filePath = $projectRootPath . '/' . $filePath;
                }

                // On test si le fichier existe
                if (!file_exists($filePath)) {
                    throw new \Exception("Le fichier n'existe pas : " . $filePath);
                }

                // On charge le fichier pour hashage
                $fileContent = file_get_contents($filePath);

                // On créer un nouveau basename pour éviter les doublons (uniquement si on est pas en mode test)
                $newBasename = md5($fileContent) . '.' . pathinfo($filePath, PATHINFO_EXTENSION);
                if ($isTestMode) {
                    $newBasename = pathinfo($filePath, PATHINFO_BASENAME);
                }

                // Nouveau chemin absolu
                $newFilePath = $webPath . $newBasename;

                // On copie le fichier dans le nouveau répertoire
                if (!copy($filePath, $newFilePath)) {
                    throw new \Exception("Le fichier n'a pas pu être copié.");
                }
                // On met les droits en 777
                chmod($newFilePath, 0777);

                // On remplace le chemin dans l'attribut de l'élément par le chemin accessible depuis l'API
                $formatPath = "/^(https?|ftp|s3):\/\/(localhost|[\da-z\.-]+)(\.[a-z\.]{2,6})?(:\d{2,5})?([\/\w \.-]*)*\/?$/";
                $newRelativePath = $webPathUrl . $newBasename;
                if (!preg_match($formatPath, $newRelativePath)) {
                    throw new \Exception("L'url n'est pas valide pour : " . $newRelativePath);
                }
                $element->setAttribute($attribute, $newRelativePath);
            }
        }

        // On sauvegarde le nouveau contenu HTML 
        $htmlContent = $dom->saveHTML();

        // Charge les options du module
        $styleAttributes = '';
        if ($pdfOptions['noBreakTableRows']) {
            $styleAttributes .= 'tr, td, th { page-break-inside: avoid; }';
        }
        if (isset($pdfOptions['font'])) {
            $styleAttributes .= 'html { ';
            foreach ($pdfOptions['font'] as $font => $att) {
                $styleAttributes .= 'font-' . $font . ': ' . $att . '; ';
            }
            $styleAttributes .= '}';
        }

        // Si pas d'options de module on retourne le contenu HTML
        if (empty($styleAttributes)) {
            return $htmlContent;
        }

        // Recharge le contenu HTML
        $dom->loadHTML($htmlContent);

        // Si options module on ajoute le style dans le head
        if (!empty($styleAttributes) && $dom->getElementsByTagName('style')->length === 0) { // Si pas de balise style on la créer
            $style = $dom->createElement('style', $styleAttributes);
        }
        if (!empty($styleAttributes) && $dom->getElementsByTagName('style')->length > 0) { // Si balise style on ajoute les attributs
            $style = $dom->getElementsByTagName('style')->item(0);
            $style->nodeValue .= $styleAttributes;
        }
        if (!empty($styleAttributes) && $dom->getElementsByTagName('head')->length > 0) { // Si balise head on ajoute la balise style
            $dom->getElementsByTagName('head')->item(0)->appendChild($style);
        }
        if (!empty($styleAttributes) && $dom->getElementsByTagName('head')->length === 0) { // Si pas de balise head on la créer
            $head = $dom->createElement('head');
            if ($dom->getElementsByTagName('html')->item(0)->hasChildNodes()) {
                $firstChild = $dom->getElementsByTagName('html')->item(0)->firstChild;
                $dom->getElementsByTagName('html')->item(0)->insertBefore($head, $firstChild);
            } else {
                $dom->getElementsByTagName('html')->item(0)->appendChild($head);
            }
            $dom->getElementsByTagName('head')->item(0)->appendChild($style);
        }

        return $dom->saveHTML();
    }

    /**
     * Supprime les images temporaires
     * 
     * @param string $webPath Chemin vers le répertoire web (ex: /var/www/html/public-data/tmp/url_to_pdf/)
     * @return void
     */
    public function deleteTempImages(string $webPath): void
    {
        // Test si $webPath se termine par un slash, si non on le rajoute
        if (substr($webPath, -1) !== '/') {
            $webPath .= '/';
        }

        // Check si $webPath se termine par /pdf/ si oui on supprime le /pdf/
        if (substr($webPath, -5) === '/pdf/') {
            $webPath = substr($webPath, 0, -4);
        }

        // On créer le répertoire s'il n'existe pas
        if (!is_dir($webPath)) {
            return;
        }

        // On supprime les images du répertoire
        $extensions = ['jpg', 'jpeg', 'png', 'gif'];
        foreach ($extensions as $extension) {
            array_map('unlink', glob($webPath . '*.' . $extension));
        }
    }

    /**
     * Supprime un dossier et son contenu m^me si non vide
     * 
     * @param string $dir Chemin vers le répertoire à supprimer
     * @return bool (true si le dossier a été supprimé)
     */
    public function deleteDir(string $dir): bool
    {
        if (!is_dir($dir)) {
            return false;
        }

        $files = array_diff(scandir($dir), ['.', '..']);
        foreach ($files as $file) {
            (is_dir("$dir/$file")) ? $this->deleteDir("$dir/$file") : unlink("$dir/$file");
        }

        return rmdir($dir);
    }
}
